from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from enum import auto, IntEnum
from functools import partial
from importlib import import_module
from json import dump, dumps, load, loads
from math import inf
from typing import Any

from ordered_set import OrderedSet
import numpy as np
import numpy.linalg as LA


def euclidean_projection_on_probability_simplex(input_):
    """Euclidean projection of the input on a probability simplex.

    >>> euclidean_projection_on_probability_simplex(np.array([0.2, 0.5, 0.3]))
    array([0.2, 0.5, 0.3])
    >>> euclidean_projection_on_probability_simplex(np.array([0.2, -0.3, 2]))
    array([0., 0., 1.])
    >>> euclidean_projection_on_probability_simplex(np.array([5, 5, 5, 5]))
    array([0.25, 0.25, 0.25, 0.25])
    >>> euclidean_projection_on_probability_simplex(np.array([10, 0, 0]))
    array([1., 0., 0.])
    >>> euclidean_projection_on_probability_simplex(np.array([0.6]))
    array([1.])
    >>> euclidean_projection_on_probability_simplex(np.array([0, 0, 0, 0, 0]))
    array([0.2, 0.2, 0.2, 0.2, 0.2])

    :param input_: The input to be projected.
    :return: The projection output.
    """
    sorted_input = np.flip(np.sort(input_))
    cumsum_sorted_input = sorted_input.cumsum()
    indices = np.arange(1, input_.size + 1)
    conditions = (sorted_input + (1 - cumsum_sorted_input) / indices) > 0
    rho = np.where(conditions)[0].max() + 1
    lambda_ = (1 - cumsum_sorted_input[rho - 1]) / rho
    output = (input_ + lambda_).clip(0)

    return output


def stationary_distribution(stochastic_matrix):
    P = stochastic_matrix

    if not np.allclose(P.sum(1), 1):
        raise ValueError('matrix not stochastic')

    eigenvalues, eigenvectors = LA.eig(P.T)
    pi = eigenvectors[:, np.isclose(eigenvalues, 1)][:, 0]
    pi /= pi.sum()
    pi = pi.real

    assert np.allclose(pi @ P, pi)

    return pi


def import_string(dotted_path):
    """Import an object from a module.

    >>> import_string('math.inf')
    inf

    :param dotted_path: The dotted path of the object to import.
    :return: Imported object.
    """
    module_path, class_name = dotted_path.rsplit('.', 1)

    return getattr(import_module(module_path), class_name)


def split(values, counts):
    """Split concatenated values.

    >>> split([0, 1, 2, 3, 4, 5], [3, 0, 1, 2])
    [[0, 1, 2], [], [3], [4, 5]]

    :param values: Values to be split.
    :param counts: The size of the partitions.
    :return: The split values.
    """
    splits = []
    begin = 0

    for count in counts:
        end = begin + count

        splits.append(values[begin:end])

        begin = end

    return splits


class Serializable(ABC):
    @classmethod
    @abstractmethod
    def deserialize(cls, raw_data):
        pass

    @classmethod
    def load(cls, *args, **kwargs):
        return cls.deserialize(load(*args, **kwargs))

    @classmethod
    def loads(cls, *args, **kwargs):
        return cls.deserialize(loads(*args, **kwargs))

    @abstractmethod
    def serialize(self):
        pass

    def dump(self, *args, **kwargs):
        return dump(self.serialize(), *args, **kwargs)

    def dumps(self, *args, **kwargs):
        return dumps(self.serialize(), *args, **kwargs)


@dataclass
class TreeFormSequentialDecisionProcess(Serializable):
    class NodeType(IntEnum):
        DECISION_POINT = auto()
        OBSERVATION_POINT = auto()
        END_OF_THE_DECISION_PROCESS = auto()

    @classmethod
    def deserialize_all(cls, raw_data):
        return list(map(cls.deserialize, raw_data))

    @classmethod
    def deserialize(cls, raw_data):
        transitions = {}
        node_types = {}

        for raw_transition in raw_data:
            parent_edge = tuple(raw_transition['parent_edge'])
            node_id = raw_transition['node']['id']
            node_type = getattr(cls.NodeType, raw_transition['node']['type'])
            transitions[parent_edge] = node_id
            node_types[node_id] = node_type

        return cls(transitions, node_types)

    transitions: Any
    """Entries are assumed be topologically sorted."""
    node_types: Any
    nodes: Any = field(init=False, default_factory=OrderedSet)
    decision_points: Any = field(init=False, default_factory=OrderedSet)
    observation_points: Any = field(init=False, default_factory=OrderedSet)
    sequences: Any = field(init=False, default_factory=OrderedSet)
    indices: Any = field(init=False, default_factory=dict)
    parent_sequences: Any = field(init=False, default_factory=dict)
    actions: Any = field(
        init=False,
        default_factory=partial(defaultdict, OrderedSet),
    )
    signals: Any = field(
        init=False,
        default_factory=partial(defaultdict, OrderedSet),
    )

    def __post_init__(self):
        self.nodes.update(self.transitions.values())

        for parent_edge, p in self.transitions.items():
            match self.node_types[p]:
                case self.NodeType.DECISION_POINT:
                    self.decision_points.add(p)
                case self.NodeType.OBSERVATION_POINT:
                    self.observation_points.add(p)
                case self.NodeType.END_OF_THE_DECISION_PROCESS:
                    pass
                case _:
                    raise ValueError('unknown node type')

            if parent_edge:
                parent, event = parent_edge

                match self.node_types[parent]:
                    case self.NodeType.DECISION_POINT:
                        is_sequence = True
                        parent_sequence = parent_edge

                        self.actions[parent].add(event)
                    case self.NodeType.OBSERVATION_POINT:
                        is_sequence = False
                        parent_sequence = self.parent_sequences[parent]

                        self.signals[parent].add(event)
                    case self.NodeType.END_OF_THE_DECISION_PROCESS:
                        raise ValueError('parent is an end of the tfsdp')
                    case _:
                        raise ValueError('unknown parent node type')
            else:
                is_sequence = True
                parent_sequence = parent_edge

            if is_sequence:
                self.sequences.add(parent_edge)

                self.indices[parent_edge] = len(self.indices)

            self.parent_sequences[p] = parent_sequence

    def behavioral_uniform_strategy(self):
        strategy = {
            j: np.full(len(self.actions[j]), 1 / len(self.actions[j]))
            for j in self.decision_points
        }

        return strategy

    def behavioral_best_response(self, utility):
        strategy = {}
        V = defaultdict(int)

        for p in reversed(self.nodes):
            match self.node_types[p]:
                case self.NodeType.DECISION_POINT:
                    V[p] = -inf
                    index = None

                    for i, a in enumerate(self.actions[p]):
                        value = (
                            utility[self.indices[p, a]]
                            + V[self.transitions[p, a]]
                        )

                        if V[p] < value:
                            V[p] = value
                            index = i

                    strategy[p] = np.zeros(len(self.actions[p]))
                    strategy[p][index] = 1
                case self.NodeType.OBSERVATION_POINT:
                    for s in self.signals[p]:
                        V[p] += V[self.transitions[p, s]]

        return strategy, V[self.nodes[0]]

    def sequence_form_best_response(self, utility):
        strategy, value = self.behavioral_best_response(utility)

        return self.behavioral_to_sequence_form(strategy), value

    def behavioral_to_sequence_form(self, behavioral_strategy):
        strategy = np.zeros(len(self.sequences))
        strategy[self.indices[()]] = 1

        for j in self.decision_points:
            p_j = self.parent_sequences[j]

            for i, a in enumerate(self.actions[j]):
                strategy[self.indices[j, a]] = (
                    behavioral_strategy[j][i]
                    * strategy[self.indices[p_j]]
                )

        return strategy

    def counterfactual_utilities(self, behavioral_strategy, utility):
        V = defaultdict(int)

        for p in reversed(self.nodes):
            match self.node_types[p]:
                case self.NodeType.DECISION_POINT:
                    for i, a in enumerate(self.actions[p]):
                        V[p] += (
                            behavioral_strategy[p][i]
                            * (
                                utility[self.indices[p, a]]
                                + V[self.transitions[p, a]]
                            )
                        )
                case self.NodeType.OBSERVATION_POINT:
                    for s in self.signals[p]:
                        V[p] += V[self.transitions[p, s]]

        utilities = {}

        for j in self.decision_points:
            utilities[j] = np.empty(len(self.actions[j]))

            for i, a in enumerate(self.actions[j]):
                utilities[j][i] = (
                    utility[self.indices[j, a]]
                    + V[self.transitions[j, a]]
                )

        return utilities

    def serialize(self):
        raw_data = []

        for parent_edge, p in self.transitions.items():
            raw_data.append(
                {
                    'parent_edge': parent_edge,
                    'node': {'id': p, 'type': self.node_types[p].name},
                },
            )

        return raw_data



from abc import ABC, abstractmethod
from dataclasses import dataclass, field, KW_ONLY
from functools import partial
from itertools import repeat
from math import inf, isinf
from typing import Any

from scipy.special import softmax
import numpy as np

from noregret.utilities import (
    euclidean_projection_on_probability_simplex,
    split,
    stationary_distribution,
)


@dataclass
class RegretMinimizer(ABC):
    """Regret minimizer."""

    dimension: Any
    _: KW_ONLY
    gamma: Any = 0
    is_time_symmetric: bool = False
    iteration_count: Any = 0
    strategies: Any = field(default_factory=list)
    previous_strategy: Any = None
    weight_sum: Any = 0
    average_strategy: Any = None
    utilities: Any = field(default_factory=list)
    previous_utility: Any = None
    cumulative_utility: Any = None

    def __post_init__(self):
        if self.previous_strategy is None:
            self.previous_strategy = np.full(
                self.dimension,
                1 / self.dimension,
            )

        if self.average_strategy is None:
            self.average_strategy = np.full(self.dimension, 1 / self.dimension)

        if self.previous_utility is None:
            self.previous_utility = np.zeros(self.dimension)

        if self.cumulative_utility is None:
            self.cumulative_utility = np.zeros(self.dimension)

    @abstractmethod
    def next_strategy(self, prediction=False):
        pass

    def observe_utility(self, utility):
        if len(self.strategies) == len(self.utilities):
            raise ValueError('next strategy not yet outputted')
        elif len(self.strategies) != len(self.utilities) + 1:
            raise ValueError('more than one strategies were outputted')

        strategy = self.strategies[-1]
        self.iteration_count += 1

        if self.is_time_symmetric:
            self.strategies.clear()

        self.previous_strategy = strategy.copy()

        self.update_average_strategy()

        if not self.is_time_symmetric:
            self.utilities.append(utility.copy())

        self.previous_utility = utility.copy()
        self.cumulative_utility += utility

    def update_average_strategy(self):
        if self.gamma > 0 and isinf(self.gamma):
            weight = inf
        else:
            weight = self.iteration_count ** self.gamma

        self.weight_sum += weight

        if isinf(weight):
            self.average_strategy = self.previous_strategy
        else:
            self.average_strategy += (
                weight
                * (self.previous_strategy - self.average_strategy)
                / self.weight_sum
            )


@dataclass
class ProbabilitySimplexRegretMinimizer(RegretMinimizer, ABC):
    """Regret minimizer.

    Assumes a linear utility function, and optimizes over the
    probability simplex.
    """

    _: KW_ONLY
    regrets: Any = field(default_factory=list)
    previous_regrets: Any = None
    cumulative_regrets: Any = None
    cumulative_regret: Any = 0

    def __post_init__(self):
        super().__post_init__()

        if self.previous_regrets is None:
            self.previous_regrets = np.zeros(self.dimension)

        if self.cumulative_regrets is None:
            self.cumulative_regrets = np.zeros(self.dimension)

    def observe_utility(self, utility):
        super().observe_utility(utility)

        regrets = utility - (self.previous_strategy @ utility)

        if not self.is_time_symmetric:
            self.regrets.append(regrets.copy())

        self.previous_regrets = regrets.copy()
        self.cumulative_regrets += regrets
        self.cumulative_regret = self.cumulative_regrets.max()


@dataclass
class FollowTheRegularizedLeader(ProbabilitySimplexRegretMinimizer, ABC):
    """Follow the regularized leader (FTLR)."""

    learning_rate: Any


@dataclass
class MultiplicativeWeightsUpdate(FollowTheRegularizedLeader):
    """Multiplicative weights update (MWU)."""

    def next_strategy(self, prediction=False):
        if prediction is False:
            theta = self.cumulative_utility
        else:
            if prediction is True:
                prediction = self.previous_utility

            theta = prediction + self.cumulative_utility

        strategy = softmax(self.learning_rate * theta)

        self.strategies.append(strategy)

        return strategy


@dataclass
class EuclideanRegularization(FollowTheRegularizedLeader):
    """Euclidean regularization."""

    def next_strategy(self, prediction=False):
        if prediction is False:
            theta = self.cumulative_utility
        else:
            if prediction is True:
                prediction = self.previous_utility

            theta = prediction + self.cumulative_utility

        strategy = euclidean_projection_on_probability_simplex(
            self.learning_rate * theta,
        )

        self.strategies.append(strategy)

        return strategy


@dataclass
class MirrorDescent(ProbabilitySimplexRegretMinimizer, ABC):
    """(Online) mirror descent (MR)."""

    learning_rate: Any


@dataclass
class OnlineGradientDescent(MirrorDescent):
    """Online gradient descent."""

    def next_strategy(self, prediction=False):
        if prediction is False or prediction is True:
            theta = self.previous_utility
        else:
            theta = prediction

        strategy = euclidean_projection_on_probability_simplex(
            self.previous_strategy + self.learning_rate * theta,
        )

        self.strategies.append(strategy)

        return strategy


@dataclass
class RegretMatching(ProbabilitySimplexRegretMinimizer):
    """Regret matching."""

    def next_strategy(self, prediction=False):
        if prediction is False:
            theta = self.cumulative_regrets
        else:
            if prediction is True:
                prediction = self.previous_utility

            theta = (
                prediction
                - prediction @ self.previous_strategy
                + self.cumulative_regrets
            )

        return self._next_strategy(theta.clip(0))

    def _next_strategy(self, unnormalized_strategy):
        if np.allclose(unnormalized_strategy, 0):
            strategy = np.full(self.dimension, 1 / self.dimension)
        else:
            strategy = unnormalized_strategy / unnormalized_strategy.sum()

        self.strategies.append(strategy)

        return strategy


@dataclass
class RegretMatchingPlus(RegretMatching):
    """Regret matching plus."""

    _: KW_ONLY
    cumulative_regrets_plus: Any = None

    def __post_init__(self):
        super().__post_init__()

        if self.cumulative_regrets_plus is None:
            self.cumulative_regrets_plus = np.zeros(self.dimension)

    def next_strategy(self, prediction=False):
        if prediction is False:
            theta = self.cumulative_regrets_plus
        else:
            if prediction is True:
                prediction = self.previous_utility

            theta = (
                prediction
                - prediction @ self.previous_strategy
                + self.cumulative_regrets_plus
            ).clip(0)

        return self._next_strategy(theta)

    def observe_utility(self, utility):
        super().observe_utility(utility)

        self.cumulative_regrets_plus += self.previous_regrets
        self.cumulative_regrets_plus = self.cumulative_regrets_plus.clip(0)


class DiscountedStrategyAveragingMixin:
    gamma: Any
    iteration_count: Any
    previous_strategy: Any
    weight_sum: Any
    average_strategy: Any

    def update_average_strategy(self):
        T = self.iteration_count
        weight = (T / (T + 1)) ** self.gamma
        self.weight_sum += weight
        self.average_strategy += (
            weight
            * (self.previous_strategy - self.average_strategy)
            / self.weight_sum
        )


@dataclass
class DiscountedRegretMatching(
        DiscountedStrategyAveragingMixin,
        RegretMatching,
):
    """Discounted regret matching (used in DCFR)."""

    _: KW_ONLY
    alpha: Any = 0
    beta: Any = 0
    discounted_regrets: Any = None

    def __post_init__(self):
        super().__post_init__()

        if self.discounted_regrets is None:
            self.discounted_regrets = np.zeros(self.dimension)

    def next_strategy(self, prediction=False):
        if prediction is False:
            theta = self.discounted_regrets
        else:
            if prediction is True:
                prediction = self.previous_utility

            theta = (
                prediction
                - prediction @ self.previous_strategy
                + self.discounted_regrets
            )
            T = self.iteration_count + 1
            theta[theta > 0] *= T ** self.alpha / (T ** self.alpha + 1)
            theta[theta < 0] *= T ** self.beta / (T ** self.beta + 1)

        return self._next_strategy(theta.clip(0))

    def observe_utility(self, utility):
        super().observe_utility(utility)

        self.discounted_regrets += self.previous_regrets
        T = self.iteration_count
        self.discounted_regrets[self.discounted_regrets > 0] *= (
            T ** self.alpha / (T ** self.alpha + 1)
        )
        self.discounted_regrets[self.discounted_regrets < 0] *= (
            T ** self.beta / (T ** self.beta + 1)
        )


@dataclass
class SwapRegretMinimizer(RegretMinimizer, ABC):
    """Swap regret minimizer."""


@dataclass
class ProbabilitySimplexSwapRegretMinimizer(SwapRegretMinimizer, ABC):
    """Swap regret minimizer.

    Optimizes over the probability simplex.
    """


@dataclass
class BlumMansour(ProbabilitySimplexSwapRegretMinimizer):
    """Blum-Mansour algorithm.

    External regret minimizers optimize over the probability simplex.
    """

    regret_minimizer_factory: Any = None
    _: KW_ONLY
    outputs: Any = field(init=False)
    external_regret_minimizers: Any = None

    def __post_init__(self):
        super().__post_init__()

        d = self.dimension
        self.outputs = np.full((d, d), 1 / d)

        if self.external_regret_minimizers is None:
            if self.regret_minimizer_factory is None:
                raise ValueError('regret minimizer factory not set')

            self.external_regret_minimizers = tuple(
                map(self.regret_minimizer_factory, repeat(d, d)),
            )

    def next_strategy(self, prediction=False):
        if prediction is False or prediction is True:
            for a, R in enumerate(self.external_regret_minimizers):
                self.outputs[:, a] = R.next_strategy(prediction)
        else:
            for a, R in enumerate(self.external_regret_minimizers):
                self.outputs[:, a] = R.next_strategy(
                    self.previous_strategy[a] * prediction,
                )

        strategy = stationary_distribution(self.outputs.T)

        self.strategies.append(strategy)

        return strategy

    def observe_utility(self, utility):
        super().observe_utility(utility)

        for a, R in enumerate(self.external_regret_minimizers):
            R.observe_utility(self.previous_strategy[a] * utility)


@dataclass
class SequenceFormPolytopeRegretMinimizer(RegretMinimizer, ABC):
    """Regret minimizer.

    Optimizes over the sequence-form polytope.
    """


@dataclass
class CounterfactualRegretMinimization(SequenceFormPolytopeRegretMinimizer):
    """Counterfactual regret minimization (CFR).

    Local regret minimizers optimize over the probability simplex.
    """

    tree_form_sequential_decision_process: Any
    regret_minimizer_factory: Any = RegretMatching
    _: KW_ONLY
    dimension: Any = field(init=False)
    behavioral_strategy: Any = field(init=False)
    previous_behavioral_strategy: Any = None
    local_regret_minimizers: Any = None

    def __post_init__(self):
        tfsdp = self.tree_form_sequential_decision_process
        self.dimension = len(tfsdp.sequences)

        super().__post_init__()

        self.behavioral_strategy = tfsdp.behavioral_uniform_strategy()

        if self.previous_behavioral_strategy is None:
            self.previous_behavioral_strategy = self.behavioral_strategy.copy()

        if self.local_regret_minimizers is None:
            if self.regret_minimizer_factory is None:
                raise ValueError('regret minimizer factory not set')

            self.local_regret_minimizers = {
                j: self.regret_minimizer_factory(len(tfsdp.actions[j]))
                for j in tfsdp.decision_points
            }

    def next_strategy(self, prediction=False):
        self.behavioral_strategy = {}

        if prediction is False or prediction is True:
            for j, R in self.local_regret_minimizers.items():
                self.behavioral_strategy[j] = R.next_strategy(prediction)
        else:
            counterfactual_predictions = (
                self
                .tree_form_sequential_decision_process
                .counterfactual_utilities(
                    self.previous_behavioral_strategy,
                    prediction,
                )
            )

            for j, R in self.local_regret_minimizers.items():
                self.behavioral_strategy[j] = R.next_strategy(
                    counterfactual_predictions[j],
                )

        strategy = (
            self
            .tree_form_sequential_decision_process
            .behavioral_to_sequence_form(self.behavioral_strategy)
        )

        self.strategies.append(strategy)

        return strategy

    def observe_utility(self, utility):
        super().observe_utility(utility)

        counterfactual_utilities = (
            self
            .tree_form_sequential_decision_process
            .counterfactual_utilities(self.behavioral_strategy, utility)
        )

        for j, R in self.local_regret_minimizers.items():
            R.observe_utility(counterfactual_utilities[j])

        self.previous_behavioral_strategy = self.behavioral_strategy.copy()


@dataclass
class CounterfactualRegretMinimizationPlus(CounterfactualRegretMinimization):
    """Counterfactual regret minimization plus (CFR+)."""

    regret_minimizer_factory: Any = RegretMatchingPlus
    _: KW_ONLY
    gamma: Any = 1


@dataclass
class DiscountedCounterfactualRegretMinimization(
        DiscountedStrategyAveragingMixin,
        CounterfactualRegretMinimization,
):
    """Discounted counterfactual regret minimization (DCFR)."""

    regret_minimizer_factory: Any = None
    _: KW_ONLY
    alpha: Any = 1.5
    beta: Any = 0
    gamma: Any = 2

    def __post_init__(self):
        if self.regret_minimizer_factory is None:
            self.regret_minimizer_factory = partial(
                DiscountedRegretMatching,
                alpha=self.alpha,
                beta=self.beta,
                gamma=self.gamma,
            )

        super().__post_init__()


@dataclass
class RegretCircuit(RegretMinimizer):
    """Regret circuit.

    Assumes a linear utility function.
    """


@dataclass
class CartesianProductRegretCircuit(RegretCircuit):
    """Regret circuit for the cartesian product operation.

    Optimizes over the cartesian product of sets, over which the components
    optimize.
    """

    component_regret_minimizers: Any
    _: KW_ONLY
    dimension: Any = field(init=False)

    def __post_init__(self):
        self.dimension = sum(self.dimensions)

        super().__post_init__()

    @property
    def dimensions(self):
        return np.array(
            [R.dimension for R in self.component_regret_minimizers],
        )

    def next_strategy(self, prediction=False):
        if prediction is False or prediction is True:
            predictions = repeat(prediction)
        else:
            predictions = split(prediction, self.dimensions)

        strategy = []

        for R, m in zip(self.component_regret_minimizers, predictions):
            strategy.append(R.next_strategy(m))

        strategy = np.concatenate(strategy)

        self.strategies.append(strategy)

        return strategy

    def observe_utility(self, utility):
        super().observe_utility(utility)

        for R, utility in zip(
                self.component_regret_minimizers,
                split(utility, self.dimensions),
        ):
            R.observe_utility(utility)


@dataclass
class ConvexHullRegretCircuit(RegretCircuit):
    """Regret circuit for the convex hull operation.

    Optimizes over the convex hull of sets, over which the components
    optimize.
    """

    component_regret_minimizers: Any
    regret_minimizer_factory: Any = None
    _: KW_ONLY
    dimension: Any = field(init=False)
    outputs: Any = field(init=False)
    previous_outputs: Any = None
    mixing_regret_minimizer: Any = None

    def __post_init__(self):
        if self.component_regret_minimizers:
            self.dimension = self.component_regret_minimizers[0].dimension
        else:
            self.dimension = 0

        super().__post_init__()

        self.outputs = np.full(
            (len(self.component_regret_minimizers), self.dimension),
            1 / self.dimension,
        )

        if self.previous_outputs is None:
            self.previous_outputs = self.outputs.copy()

        if self.mixing_regret_minimizer is None:
            if self.regret_minimizer_factory is None:
                raise ValueError('regret minimizer factory not set')

            self.mixing_regret_minimizer = self.regret_minimizer_factory(
                len(self.component_regret_minimizers),
            )

    def next_strategy(self, prediction=False):
        for i, R in enumerate(self.component_regret_minimizers):
            self.outputs[i] = R.next_strategy(prediction)

        if prediction is False or prediction is True:
            m = prediction
        else:
            m = self.previous_outputs @ prediction

        probabilities = self.mixing_regret_minimizer.next_strategy(m)
        strategy = self.outputs.T @ probabilities

        self.strategies.append(strategy)

        return strategy

    def observe_utility(self, utility):
        super().observe_utility(utility)

        for R in self.component_regret_minimizers:
            R.observe_utility(utility)

        self.previous_outputs = self.outputs.copy()

        self.mixing_regret_minimizer.observe_utility(self.outputs @ utility)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import partial
from itertools import count, permutations
from math import factorial
from typing import Any

from scipy.sparse import lil_array
import numpy as np

from noregret.utilities import (
    Serializable,
    split,
    TreeFormSequentialDecisionProcess,
)


@dataclass
class Game(ABC):
    """Game."""

    def __post_init__(self):
        self._verify()

    def _verify(self, **kwargs):
        pass

    @property
    @abstractmethod
    def player_count(self):
        pass

    @abstractmethod
    def dimension(self, player):
        pass

    @property
    def dimensions(self):
        return np.array(list(map(self.dimension, range(self.player_count))))

    @abstractmethod
    def utility(self, player, *opponent_strategies):
        pass

    @abstractmethod
    def value(self, player, *strategies):
        pass

    def values(self, *strategies):
        return np.array(
            [self.value(i, *strategies) for i in range(self.player_count)],
        )

    @abstractmethod
    def correlated_value(self, player, *strategies):
        pass

    def correlated_values(self, *strategies):
        return np.array(
            [
                self.correlated_value(i, *strategies)
                for i in range(self.player_count)
            ],
        )

    @abstractmethod
    def best_response(self, player, *opponent_strategies):
        pass

    def nash_gap(self, *strategies):
        gap = 0

        for i, value in enumerate(self.values(*strategies)):
            opponent_strategies = strategies[:i] + strategies[i + 1:]
            _, br_value = self.best_response(i, *opponent_strategies)
            gap += br_value - value

        return gap

    def cce_gap(self, *strategies):
        average_strategies = list(map(partial(np.mean, axis=0), strategies))
        gap = 0

        for i, value in enumerate(self.correlated_values(*strategies)):
            average_opponent_strategies = (
                average_strategies[:i] + average_strategies[i + 1:]
            )
            _, br_value = self.best_response(i, *average_opponent_strategies)
            gap += br_value - value

        return gap


@dataclass
class TwoPlayerGame(Game, ABC):
    """Two-player (2p) game.

    Row and column players are of indices 0 and 1, respectively.
    """

    def _verify(self, **kwargs):
        super()._verify(**kwargs)

        if self.player_count != 2:
            raise ValueError('number of players not 2')

    @property
    @abstractmethod
    def row_utilities(self):
        pass

    @property
    @abstractmethod
    def column_utilities(self):
        pass

    def dimension(self, player):
        match player:
            case 0:
                dimension = self.row_dimension
            case 1:
                dimension = self.column_dimension
            case _:
                raise ValueError(f'Player {player} does not exist')

        return dimension

    @property
    def row_dimension(self):
        return self.row_utilities.shape[0]

    @property
    def column_dimension(self):
        return self.row_utilities.shape[1]

    def utility(self, player, opponent_strategy):
        match player:
            case 0:
                utility = self.row_utility(opponent_strategy)
            case 1:
                utility = self.column_utility(opponent_strategy)
            case _:
                raise ValueError(f'Player {player} does not exist')

        return utility

    def row_utility(self, column_strategy):
        return self.row_utilities @ column_strategy

    def column_utility(self, row_strategy):
        return row_strategy @ self.column_utilities

    def value(self, player, row_strategy, column_strategy):
        match player:
            case 0:
                value = self.row_value(row_strategy, column_strategy)
            case 1:
                value = self.column_value(row_strategy, column_strategy)
            case _:
                raise ValueError(f'Player {player} does not exist')

        return value

    def row_value(self, row_strategy, column_strategy):
        return row_strategy @ self.row_utilities @ column_strategy

    def column_value(self, row_strategy, column_strategy):
        return row_strategy @ self.column_utilities @ column_strategy

    def correlated_value(self, player, row_strategies, column_strategies):
        match player:
            case 0:
                value = self.correlated_row_value(
                    row_strategies,
                    column_strategies,
                )
            case 1:
                value = self.correlated_column_value(
                    row_strategies,
                    column_strategies,
                )
            case _:
                raise ValueError(f'Player {player} does not exist')

        return value

    def correlated_row_value(self, row_strategies, column_strategies):
        return (
            row_strategies @ self.row_utilities * column_strategies
        ).sum(1).mean()

    def correlated_column_value(self, row_strategies, column_strategies):
        return (
            row_strategies @ self.column_utilities * column_strategies
        ).sum(1).mean()

    def best_response(self, player, opponent_strategy):
        match player:
            case 0:
                best_response = self.row_best_response(opponent_strategy)
            case 1:
                best_response = self.column_best_response(opponent_strategy)
            case _:
                raise ValueError(f'Player {player} does not exist')

        return best_response

    @abstractmethod
    def row_best_response(self, column_strategy):
        pass

    @abstractmethod
    def column_best_response(self, row_strategy):
        pass


@dataclass
class TwoPlayerZeroSumGame(TwoPlayerGame, ABC):
    """Two-player zero-sum (2p0s) game."""

    @property
    def column_utilities(self):
        return -self.row_utilities

    def values(self, row_strategy, column_strategy):
        value = self.row_value(row_strategy, column_strategy)

        return np.array((value, -value))

    def correlated_values(self, row_strategies, column_strategies):
        value = self.correlated_row_value(row_strategies, column_strategies)

        return np.array((value, -value))

    def nash_gap(self, row_strategy, column_strategy):
        _, row_best_response_value = self.row_best_response(column_strategy)
        _, column_best_response_value = self.column_best_response(row_strategy)

        return row_best_response_value + column_best_response_value

    def exploitability(self, row_strategy, column_strategy):
        return self.nash_gap(row_strategy, column_strategy) / 2


@dataclass
class NormalFormGame(Serializable, Game):
    """Normal-form game.

    Each player optimizes over the probability simplex.
    """

    @classmethod
    def deserialize(cls, raw_data):
        return cls(raw_data['actions'], np.array(raw_data['utilities']))

    actions: Any
    utilities: Any
    indices: Any = field(init=False, default_factory=list)

    def __post_init__(self):
        super().__post_init__()

        for i, actions in enumerate(self.actions):
            self.indices.append(dict(zip(actions, count())))

    def _verify(self, *, utilities_shape=None, **kwargs):
        super()._verify(**kwargs)

        if utilities_shape is None:
            utilities_shape = (*map(len, self.actions), self.player_count)

        if self.utilities.shape != utilities_shape:
            raise ValueError('utilities do not match actions and players')

    @property
    def player_count(self):
        return len(self.actions)

    def utility(self, player, *opponent_strategies):
        raise NotImplementedError

    def value(self, player, *strategies):
        raise NotImplementedError

    def correlated_value(self, player, *strategies):
        raise NotImplementedError

    def best_response(self, player, *opponent_strategies):
        raise NotImplementedError

    def serialize(self):
        return {'actions': self.actions, 'utilities': self.utilities.tolist()}


@dataclass
class TwoPlayerNormalFormGame(TwoPlayerGame, NormalFormGame):
    """Two-player (2p) normal-form game."""

    @property
    def row_actions(self):
        return self.actions[0]

    @property
    def column_actions(self):
        return self.actions[1]

    @property
    def row_indices(self):
        return self.indices[0]

    @property
    def column_indices(self):
        return self.indices[1]

    @property
    def row_utilities(self):
        return self.utilities[:, :, 0]

    @property
    def column_utilities(self):
        return self.utilities[:, :, 1]

    def row_best_response(self, column_strategy):
        strategy = np.zeros(len(self.row_actions))
        utility = self.row_utility(column_strategy)
        index = utility.argmax()
        strategy[index] = 1

        return strategy, utility[index]

    def column_best_response(self, row_strategy):
        strategy = np.zeros(len(self.column_actions))
        utility = self.column_utility(row_strategy)
        index = utility.argmax()
        strategy[index] = 1

        return strategy, utility[index]


@dataclass
class TwoPlayerZeroSumNormalFormGame(
        TwoPlayerZeroSumGame,
        TwoPlayerNormalFormGame,
):
    """Two-player zero-sum (2p0s) normal-form game.

    The utility matrix is from the viewpoint of the row player.
    """

    def _verify(self, **kwargs):
        super()._verify(
            **kwargs,
            utilities_shape=(len(self.row_actions), len(self.column_actions)),
        )

    @property
    def row_utilities(self):
        return self.utilities


@dataclass
class ExtensiveFormGame(Serializable, Game):
    """Extensive-form game (EFG).

    Each player optimizes over the sequence-form polytope.
    """

    @classmethod
    def deserialize(cls, raw_data):
        raise NotImplementedError

    tree_form_sequential_decision_processes: Any
    utilities: Any

    @property
    def player_count(self):
        return len(self.tree_form_sequential_decision_processes)

    def utility(self, player, *opponent_strategies):
        raise NotImplementedError

    def value(self, player, *strategies):
        raise NotImplementedError

    def correlated_value(self, player, *strategies):
        raise NotImplementedError

    def best_response(self, player, *opponent_strategies):
        raise NotImplementedError

    def serialize(self):
        raise NotImplementedError


@dataclass
class TwoPlayerExtensiveFormGame(TwoPlayerGame, ExtensiveFormGame):
    """Two-player (2p) extensive-form game (EFG)."""

    @classmethod
    def deserialize(cls, raw_data):
        tfsdps = TreeFormSequentialDecisionProcess.deserialize_all(
            raw_data['tree_form_sequential_decision_processes'],
        )
        shape = tuple(len(tfsdp.sequences) for tfsdp in tfsdps)
        row_utilities = lil_array(shape)
        column_utilities = lil_array(shape)

        for raw_utility in raw_data['utilities']:
            if len(raw_utility['values']) != 2:
                raise ValueError('utility is not of a 2-player game')

            indices = []

            for tfsdp, sequence in zip(tfsdps, raw_utility['sequences']):
                sequence = tuple(sequence)

                indices.append(tfsdp.indices[sequence])

            indices = tuple(indices)
            row_utilities[indices] = raw_utility['values'][0]
            column_utilities[indices] = raw_utility['values'][1]

        return cls(tfsdps, [row_utilities.tocsr(), column_utilities.tocsr()])

    def _verify(self, **kwargs):
        super()._verify(**kwargs)

        if not (
                self.row_utilities.shape
                == self.column_utilities.shape
                == (len(self.row_sequences), len(self.column_sequences))
        ):
            raise ValueError('utilities do not match sequences')

    @property
    def row_tree_form_sequential_decision_process(self):
        return self.tree_form_sequential_decision_processes[0]

    @property
    def column_tree_form_sequential_decision_process(self):
        return self.tree_form_sequential_decision_processes[1]

    @property
    def row_sequences(self):
        return self.row_tree_form_sequential_decision_process.sequences

    @property
    def column_sequences(self):
        return self.column_tree_form_sequential_decision_process.sequences

    @property
    def row_indices(self):
        return self.row_tree_form_sequential_decision_process.indices

    @property
    def column_indices(self):
        return self.column_tree_form_sequential_decision_process.indices

    @property
    def row_utilities(self):
        return self.utilities[0]

    @property
    def column_utilities(self):
        return self.utilities[1]

    def row_best_response(self, column_strategy):
        best_response = (
            self
            .row_tree_form_sequential_decision_process
            .sequence_form_best_response(self.row_utility(column_strategy))
        )

        return best_response

    def column_best_response(self, row_strategy):
        best_response = (
            self
            .column_tree_form_sequential_decision_process
            .sequence_form_best_response(self.column_utility(row_strategy))
        )

        return best_response

    def serialize(self):
        tfsdps = self.tree_form_sequential_decision_processes
        raw_tfsdps = [tfsdp.to_list() for tfsdp in tfsdps]
        raw_utilities = []
        abs_utility_sums = abs(self.row_utilities) + abs(self.column_utilities)

        for indices in zip(*abs_utility_sums.nonzero()):
            sequences = []

            for tfsdp, index in zip(tfsdps, indices):
                sequences.append(tfsdp.sequences[index])

            row_value = self.row_utilities[indices].item()
            column_value = self.column_utilities[indices].item()
            values = row_value, column_value

            raw_utilities.append({'sequences': sequences, 'values': values})

        return {
            'tree_form_sequential_decision_processes': raw_tfsdps,
            'utilities': raw_utilities,
        }


@dataclass
class TwoPlayerZeroSumExtensiveFormGame(
        TwoPlayerZeroSumGame,
        TwoPlayerExtensiveFormGame,
):
    """Two-player zero-sum (2p0s) extensive-form game (EFG).

    The utility matrix is from the viewpoint of the row player.
    """

    @classmethod
    def deserialize(cls, raw_data):
        tfsdps = TreeFormSequentialDecisionProcess.deserialize_all(
            raw_data['tree_form_sequential_decision_processes'],
        )
        shape = tuple(len(tfsdp.sequences) for tfsdp in tfsdps)
        utilities = lil_array(shape)

        for raw_utility in raw_data['utilities']:
            indices = []

            for tfsdp, sequence in zip(tfsdps, raw_utility['sequences']):
                sequence = tuple(sequence)

                indices.append(tfsdp.indices[sequence])

            indices = tuple(indices)
            utilities[indices] = raw_utility['value']

        return cls(tfsdps, utilities.tocsr())

    @property
    def row_utilities(self):
        return self.utilities

    def serialize(self):
        tfsdps = self.tree_form_sequential_decision_processes
        raw_tfsdps = [tfsdp.to_list() for tfsdp in tfsdps]
        raw_utilities = []

        for indices in zip(*self.utilities.nonzero()):
            sequences = []

            for tfsdp, index in zip(tfsdps, indices):
                sequences.append(tfsdp.sequences[index])

            value = self.utilities[indices].item()

            raw_utilities.append({'sequences': sequences, 'value': value})

        return {
            'tree_form_sequential_decision_processes': raw_tfsdps,
            'utilities': raw_utilities,
        }


@dataclass
class SymmetrizedGame(Game):
    """Symmetrized game.

    Each player optimizes over the cartesian product of probability
    simplices.
    """

    game: Any

    @property
    def player_count(self):
        return self.game.player_count

    def dimension(self, player):
        return sum(self.game.dimensions)

    def utility(self, player, *opponent_strategies):
        strategies = []

        for opponent_strategy in opponent_strategies:
            strategies.append(split(opponent_strategy, self.game.dimensions))

        strategies.insert(player, None)

        utilities = [0] * self.player_count

        for permutation in permutations(range(self.player_count)):
            utilities[permutation.index(player)] += self.game.utility(
                permutation.index(player),
                *(
                    strategies[permutation[i]][i]
                    for i in range(self.player_count)
                    if permutation[i] != player
                ),
            )

        utility = np.concatenate(utilities)
        utility /= factorial(self.player_count)

        return utility

    def value(self, player, *strategies):
        strategies = list(strategies)

        for i in range(self.player_count):
            strategies[i] = split(strategies[i], self.game.dimensions)

        value = 0

        for permutation in permutations(range(self.player_count)):
            value += self.game.value(
                permutation.index(player),
                *(
                    strategies[permutation[i]][i]
                    for i in range(self.player_count)
                ),
            )

        value /= factorial(self.player_count)

        return value

    def correlated_value(self, player, *strategies):
        raise NotImplementedError

    def best_response(self, player, *opponent_strategies):
        raise NotImplementedError
